import numpy as np


class Shading:
    """Parameters of the shaders."""

    def __init__(self, bg_color: np.ndarray, ambient_light_color: np.ndarray,
                 directional_light_color: np.ndarray,
                 directional_light_dir: np.ndarray):
        """Constructor.

        Args:
            bg_color (np.ndarray, optional): background color. Defaults to Black.
            ambient_light_color (np.ndarray, optional): color of ambient light. Defaults to [0.7, 0.7, 0.7].
            directional_light_color (np.ndarray, optional): color of directional light. Defaults to [0.3, 0.3, 0.3].
            directional_light_dir (np.ndarray, optional): direction of directional light. Defaults to [2, 1, 2].
        """

    @property
    def bg_color(self) -> np.ndarray:
        """The background color."""

    @property
    def ambient_light_color(self) -> np.ndarray:
        """The color of ambient light."""

    @property
    def directional_light_color(self) -> np.ndarray:
        """The directional light color."""

    @property
    def directional_light_dir(self) -> np.ndarray:
        """The direction of the directional light."""



import numpy as np


class LoopSubdivisionStencil:
    """This class enables efficient Loop subdivision of triangle meshes.

    The subdivision stencil specifies what linear combinations of existing vertices are needed, and
    stencils can be reused across meshes with different vertex positions but identical triangle topologies.
    Thus, the constructor only takes the triangles, and caches the stencil.
    The stencil can be applied to a set of vertices using the apply() method.
    """

    def apply(self, vertices: np.ndarray) -> np.ndarray:
        """Applies the stencil to a vertex buffer.

        Description:
            This method takes a set of vertices and subdivides them (using the stencil) to
            create the vertices of a more detailed mesh.

        Args:
            vertices (np.ndarray):  a (N, 3) buffer of vertices

        Returns:
            np.ndarray: a (S, 3) buffer of subdivided vertices
        """

    @property
    def triangles(self) -> np.ndarray:
        """The triangles of the subdivided mesh"""

    @property
    def vertex_count(self) -> int:
        """The number of vertices in the subdivided mesh."""

    @property
    def triangle_count(self) -> int:
        """The number of triangles in the subdivided mesh."""

    @staticmethod
    def create(triangles: np.ndarray, steps: int,
               project_to_limit: bool) -> "LoopSubdivisionStencil":
        """Creates a new stencil by subdividing the provided triangles.

        Args:
            triangles (np.ndarray): the initial triangle indices
            steps (int): specifies how many steps of subdivision to apply. Defaults to 1.
            project_to_limit (bool): specifies whether the vertices should be projected onto the
                                     limit surface in the final step of subdivision. Defaults to false.

        Return:
            LoopSubdivisionStencil
        """



from typing import Mapping, Union

import numpy as np
from typing import Optional

from .audio_track import AudioTrack
from .camera import Camera
from .canvas2d import Canvas2D
from .canvas3d import Canvas3D
from .drop_down_menu import DropDownMenu
from .graph import Graph
from .image import Image
from .label import Label
from .mesh import Mesh, MeshUpdate
from .text_panel import TextPanel
from .ui_parameters import UIParameters
from .video import Video



class QuantizationInfo:
    """Information about the results of quantization."""

    @property
    def keyframe_count(self) -> int:
        """The number of keyframes used."""

    @property
    def estimated_size_ratio(self) -> float:
        """The estimated size ratio after compression."""

    @property
    def mean_error(self) -> float:
        """The mean per-frame error."""

    @property
    def max_error(self) -> float:
        """The maximum per-frame error."""


class Scene:
    """Top level container representing an entire ScenePic Scene."""

    def __init__(self, scene_id: Optional[str] = None):
        """Constructor.

        Args:
            scene_id (str, optional): a unique identifier for the Scene (will be
                                      automatically populated if not provided). Defaults to None.
        """

    def grid(self, width: str, grid_template_rows: str, grid_template_columns: str) -> None:
        """Defines a CSS layout grid for the canvases of the ScenePic.

        Description:
            Should not be used in conjunction with `html_id` for individual canvases. See 
            https://developer.mozilla.org/en-US/docs/Web/CSS/CSS_Grid_Layout for more information.

        Args:
            width (str): CSS declaration for the width of the ScenePic.
            grid_template_rows (str): CSS declaration defining the rows of the ScenePic
            grid_template_columns (str): CSS declaration defining the columns of the ScenePic
        """

    def place(self, canvas_id: str, grid_row: str, grid_column: str) -> None:
        """Places a canvas at a specific place in the ScenePic grid.

        Note:
            For use with Scene.grid
        
        Args:
            canvas_id (str): The unique ID of the canvas to place
            grid_row (str): the CSS row specification for the Canvas
            grid_column (str): the CSS column specification for the Canvas
        """

    def create_canvas_3d(self, canvas_id: Optional[str] = None, width: int = 400, height: int = 400,
                         html_id: Optional[str] = None, camera: Camera = Camera(),
                         ui_parameters: UIParameters = UIParameters(), media_id: Optional[str] = None) -> Canvas3D:
        """Create a new 3D Canvas and append to the Scene.

        Args:
            canvas_id (str, optional): a unique identifier for the Canvas (will be
                                        automatically populated if not provided). Defaults to None.
            width (int, optional): the width in pixels of the Canvas on the HTML page. Defaults to 400.
            height (int, optional): the height in pixels of the Canvas on the HTML page. Defaults to 400.
            html_id (str, optional): id of an HTML element to use as this
                                        Canvas's parent in the HTML DOM (otherwise simply
                                        appended to document). Defaults to None.
            camera (Camera, optional): Camera parameters for the Canvas. See `Camera` for defaults.
            shading (Shading, optional): Set of Shading parameters for the Canvas. See `Shading` for defaults.
            ui_parameters (UIParameters, optional): Set of user interface parameters. See `UIParameters` for
                                                    defaults.
            media_id (str, optional): optional ID of a media file to attach to the canvas. This file will be
                                        used to drive playback, i.e. frames will be displayed in time with
                                        the playback of the media file.

        Returns:
            Canvas3D: a reference to the new Canvas object
        """

    def create_canvas_2d(self, canvas_id: Optional[str] = None, width: int = 400, height: int = 400,
                         html_id: Optional[str] = None, background_color: np.ndarray = (0, 0, 0),
                         media_id: Optional[str] = None) -> Canvas2D:
        """Create a new 2D Canvas and append to the Scene.

        Args:
            canvas_id (str, optional): a unique identifier for the Canvas (will be
                                        automatically populated if not provided). Defaults to None.
            width (int, optional): the width in pixels of the Canvas on the HTML page. Defaults to 400.
            height (int, optional): the height in pixels of the Canvas on the HTML page. Defaults to 400.
            html_id (str, optional): id of an HTML element to use as this
                                        Canvas's parent in the HTML DOM (otherwise simply
                                        appended to document). Defaults to None.
            background_color (np.ndarray, optional): the background color of the canvas. Defaults to Black.
            media_id (str, optional): optional ID of a media file to attach to the canvas. This file will be
                                        used to drive playback, i.e. frames will be displayed in time with
                                        the playback of the media file.

        Returns:
            Canvas2D: a reference to the new Canvas object
        """

    def create_graph(self, canvas_id: Optional[str] = None, width: int= 400, height: int = 400,
                     name_align: str = "left" , value_align: Optional[str] = "right",
                     html_id: Optional[str] = None, background_color: np.ndarray = (255, 255, 255),
                     margin: Graph.Margin = 10 , font_family: str = "sans-serif",
                     name_size: float = 12.0, value_size: float = 12.0,
                     media_id: Optional[str] = None) -> Graph:
        """Create a new Graph canvas and append to the Scene.

        Args:
            canvas_id (str, optional): a unique identifier for the Canvas (will be
                                        automatically populated if not provided). Defaults to None.
            width (int, optional): the width in pixels of the Canvas on the HTML page. Defaults to 400.
            height (int, optional): the height in pixels of the Canvas on the HTML page. Defaults to 400.
            name_align (str, optional): How to align the sparkline label
                                        (one of 'left', 'right', 'top', or 'bottom').
                                        Defaults to 'left'.
            value_align (str, optional): How to align the sparkline value
                                        (one of 'left', 'right', 'top', or 'bottom').
                                        Defaults to 'right'.
            html_id (str, optional): id of an HTML element to use as this
                                        Canvas's parent in the HTML DOM (otherwise simply
                                        appended to document). Defaults to None.
            background_color (np.ndarray, optional): the background color of the canvas. Defaults to White.
            margin (Margin, optional): the outer margin of the graph. Defaults to Margin(10).
            font_family (str, optional): the font family used for the graph labels. Defaults to "sans-serif".
            name_size (float, optional): the text size in pixels used for the graph labels. Defaults to 12.0.
            value_size (float, optional): the text size in pixels used for the graph values. Defaults to 12.0.
            media_id (str, optional): optional ID of a media file to attach to the canvas. This file will be
                                        used to drive playback, i.e. frames will be displayed in time with
                                        the playback of the media file.

        Returns:
            Graph: a reference to the new Graph object
        """

    def create_mesh(self, mesh_id: Optional[str] = None, layer_id: Optional[str] = None, double_sided: bool = False,
                    camera_space: bool = False, shared_color: Optional[np.ndarray] = None,
                    texture_id: Optional[Union[Image, str]] = None, nn_texture: bool = True,
                    use_texture_alpha: bool = True, is_billboard: bool = False,
                    is_label: bool = False) -> Mesh:
        """Create a new Mesh and append to the Scene.

        Description:
            Meshes are Scene-wide resources that can be reused across multiple
            Frames in multiple Canvas3Ds.

        Args:
            mesh_id (str, optional): a unique identifier for the Mesh (will be
                                    automatically populated if not provided). Defaults to None.
            layer_id (str, optional): an identifier to allow several Meshes to be grouped together
                                    into a "layer".  Layers should be used in conjunction with
                                    Canvas3D.set_layer_settings(). Defaults to None.
            double_sided (bool, optional): whether to turn off back face culling and draw the Mesh's
                                        triangles as double sided. Defaults to False.
            camera_space (bool, optional): whether this Mesh is defined in camera space (cannot
                                        be moved in the ScenePic user interface) or world space (standard).
                                        Defaults to False.
            shared_color (np.ndarray, optional): see `Mesh` constructor.
            texture_id (Union[Image, str], optional): see `Mesh` constructor.
            nn_texture (bool, optional): whether the shader should use nearest-neighbor
                        texture interpolation. Defaults to True.
            use_texture_alpha (bool, optional): whether to use the alpha channel in the texture
                                                for transparency (only relevant for textured Meshes).
                                                Defaults to True.
            is_billboard (bool, optional): draw this Mesh as a billboard (i.e. always facing the user)
                                        rather than rotating with the rest of the world. Defaults to False.
            is_label (bool, optional): this mesh will be treated specially as a label. Not for public use.
                                    Defaults to False.

        Returns:
            Mesh: a reference to the new Mesh object
        """

    def update_mesh(self, base_mesh_id: str, positions: np.ndarray,
                    normals: np.ndarray, colors: np.ndarray,
                    mesh_id: Optional[str] = None) -> MeshUpdate:
        """Create a new mesh from an existing one by replacing its vector positions, normals, and/or colors.

        Args:
            base_mesh_id (str): the unique identifier of the original base mesh. Must already have been created.
            positions (np.ndarray, optional): float32 matrix of [N, 3] new positions
            normals (np.ndarray, optional): float32 matrix of [N, 3] new normals.
            colors (np.ndarray, optional): float32 matrix of [N, 3] new colors.
            mesh_id (str, optional): a unique identifier for the new updated Mesh (will be
                                        automatically populated if not provided). Defaults to None.

        Returns:
            MeshUpdate: a reference to the MeshUpdate object
        """

    def update_instanced_mesh(self, base_mesh_id: str, positions: np.ndarray,
                              rotations: np.ndarray, colors: np.ndarray,
                              mesh_id: Optional[str] = None) -> MeshUpdate:
        """Create a new mesh from an existing one by replacing its vector positions, normals, and/or colors.

        Args:
            base_mesh_id (str): the unique identifier of the original base mesh. Must already have been created.
            positions (np.ndarray, optional): float32 matrix of [N, 3] new positions
            rotations (np.ndarray, optional): float32 matrix of [N, 4] new rotations.
            colors (np.ndarray, optional): float32 matrix of [N, 3] new colors.
            mesh_id (str, optional): a unique identifier for the new updated Mesh (will be
                                        automatically populated if not provided). Defaults to None.

        Returns:
            MeshUpdate: a reference to the MeshUpdate object
        """

    def update_mesh_positions(self, base_mesh_id: str, positions: np.ndarray,
                              mesh_id: Optional[str] = None) -> MeshUpdate:
        """Create a new mesh from an existing one by replacing its vector positions and compute normal values.

        Args:
            base_mesh_id (str): the unique identifier of the original base mesh. Must already have been created.
            positions (np.ndarray): float32 matrix of [N, 3] new positions
            mesh_id (str, optional): a unique identifier for the new updated Mesh (will be
                                        automatically populated if not provided). Defaults to None.
        
        Returns:
            MeshUpdate: a reference to the MeshUpdate object
        """

    def create_audio(self, audio_id: Optional[str] = None) -> AudioTrack:
        """Create a new AudioTrack object and append it to the Scene.

        Audio tracks are Scene-wide resources that can be attached to multiple canvases.

        Args:
            audio_id (str, optional): a unique identifier for the Audio track (will be
                                        automatically populated if not provided). Defaults to None.

        Returns:
            AudioTrack: a reference to an Audio track object
        """

    def create_video(self, video_id: Optional[str] = None) -> Video:
        """Create a new Video object and append it to the Scene.

        Videos are Scene-wide resources that can be attached to multiple canvases.

        Args:
            video_id (str, optional): a unique identifier for the Video (will be
                                        automatically populated if not provided). Defaults to None.

        Returns:
            Video: a reference to an Video object
        """

    def create_image(self, image_id: Optional[str] = None) -> Image:
        """Create a new Image and append to the Scene.

        Images are Scene-wide resources that can be reused as textures
        across multiple Meshes and Frame2Ds.

        Args:
            image_id (str, optional): a unique identifier for the Image (will be
                                        automatically populated if not provided). Defaults to None.

        Returns:
            Image: a reference to the Image object
        """

    def create_label(self, label_id:  Optional[str] = None, text: str = "Text", color: np.ndarray = (255, 255, 255),
                     size_in_pixels: float = 12.0, font_family: str = "sans-serif",
                     layer_id:  Optional[str] = None, horizontal_align: str = "left", vertical_align: str = "middle",
                     offset_distance: float = 0.1, camera_space: bool = False) -> Label:
        """Create a new Label and append to the Scene.

        Labels are Scene-wide resources that can be reused across
        multiple Meshes.

        Args:
            label_id (str, optional): a unique identifier for the Label (will be
                                        automatically populated if not provided). Defaults to None.
            text (str, optional): the text to use in the label. Defaults to "Text".
            color (np.ndarray, optional): the color to use. Defaults to White.
            size_in_pixels (float, optional): the size in pixels to display the label at. Defaults to 12.0.
            font_family (str, optional): HTML font family name. Defaults to "sans-serif".
            layer_id (str, optional): an identifier to allow several labels to be grouped together
                        into a "layer".  Layers should be used in conjunction with
                        Canvas3D.set_layer_settings(). Defaults to None.
            horizontal_align (str, optional): one of 'left', 'center', 'right'. how to align the
                                label horizontally. Defaults to "left".
            vertical_align (str, optional): one of 'top', 'middle', 'bottom'. how to align the
                            label vertically. Defaults to "middle".
            offset_distance (float, optional): distance in world space to offset the label by when
                                using left/right/top/bottom alignment. Defaults to 0.1.
            camera_space (bool, optional): whether the label should be created in
                                            camera space. Defaults to False.
        
        Returns:
            Label: a reference to the new Label object
        """

    def create_text_panel(self, text_panel_id: Optional[str] = None, html_id: Optional[str] = None,
                          add_input_box: bool = False) -> TextPanel:
        """Create a new TextPanel and add directly to the HTML page.

        Args:
            text_panel_id (str, optional): a unique identifier for the TextPanel (will be
                                            automatically populated if not provided). Defaults to None.
            html_id (str, optional): optional id of an HTML element to use as this
                                        TextPanel's parent in the HTML DOM (otherwise will
                                        simply be appended to document). Defaults to None.
            add_input_box (bool, optional): whether to add an input box to the TextPanel. Defaults to False.

        Returns:
            TextPanel: a reference to the new TextPanel object
        """

    def create_drop_down_menu(self, drop_down_menu_id: Optional[str] = None, title: Optional[str] = None,
                              html_id: Optional[str] = None) -> DropDownMenu:
        """Create a new DropDownMenu and add directly to the HTML page.

        Args:
            drop_down_menu_id (str, optional): a unique identifier for the DropDownMenu
                                                (will be automatically populated if not
                                                provided). Defaults to None.
            title (str, optional): a title for the drop down box. Defaults to None.
            html_id (str, optional): optional id of an HTML element to use as this
                                        TextPanel's parent in the HTML DOM (otherwise
                                        will simply be appended to document). Defaults to None.
        Returns:
            DropDownMenu: a reference to the new DropDownMenu object
        """

    def quantize_updates(self, relative_error_threshold: float = 1e-5,
                         absolute_error_threshold: float = -1.0,
                         base_mesh_id: Optional[str] = None,
                         per_frame_range: bool = True) -> Mapping[str, QuantizationInfo]:
        """Quantize the mesh updates.

        Description:
            Each update will be reduced in size in such a way as to minimize the expected per-value error from quantization.
            The number of keyframes (and thus the level of compression) can be controlled via two
            thresholds: the relative error threshold, and the absolute error threshold. The relative
            threshold is evaluated as a fraction of max(positions) - min(positions). The absolute
            threshold is measured in raw units.
        
            If the relative error threshold has a positive value but the absolute error threshold is negative,
            only the relative threshold will be used. Conversely, if the relative threshold is negative and the
            absolute threshold is positive then only the absolute threshold will be used. If both are positive,
            then they will both be applied. If both are negative, an exception will be raised.
        
            The quantization process will complete in such a way that (quantized - actual).abs().max() < threshold.
            In essence, each x, y, and z coordinate will individually be different by no more than this threshold
            after quantization. Thus, a possible result of quantization will that every frame is a keyframe, i.e. no
            quantization occurs. More typically, however, the algorithm finds a few "keyframe" meshes which
            minimize the expected error across the remaining (quantized) meshes.            

        Args:
            relative_error_threshold (float, optional): the maximum expected error as a multiple of the range of
                                                        values in the base mesh. Defaults to 1e-5.
            absolute_error_threshold (float, optional): the maximum expected error in absolute units. Defaults to -1.0.
            base_mesh_id (str, optional): ID of the base mesh to use as a filter on quantization. Defaults to None.
            per_frame_range (bool, optional): Whether to use the most accurate range per frame, increasing accuracy
                                                but reducing compression. Defaults to True.

        Returns:
            Mapping[str, QuantizationInfo]: information on the per-mesh quantization process
        """

    def measure_command_size(self) -> Mapping[str, int]:
        """Measures the number of bytes used by command type.

        Returns:
            Mapping[str, int]: the number of bytes per command type
        """

    @property
    def framerate(self) -> float:
        """Number of frames to display per second."""

    @property
    def status_bar_visibility(self) -> str:
        """CSS visibility for the status bar."""

    def configure_user_interface(self, ui_parameters: UIParameters) -> None:
        """Set user interface parameters across all Canvases with given UIParameters instance.

        Args:
            ui_parameters (UIParameters): global UI parameters to set
        """

    def simulate_key_presses(self, canvas_id: Union[Canvas3D, Canvas2D, str], *keypresses) -> None:
        """Simulates one or more keypresses to fire off events in the ScenePic UI at startup, e.g. video playback.

        Args:
            self (Scene): self reference
            canvas_id (Union[Canvas3D, Canvas2D, str]): ID of the canvas on which to fire off this
                                                        event.
            keypresses: the key values
        """

    def link_canvas_events(self, *canvases) -> None:
        """Specify that all input events should be linked across the provided set of Canvases.

        Args:
            self (Scene): self reference
            canvases: the canvases which should have linked events
        """

    def clear_script(self) -> None:
        """Call this if you are dumping a ScenePic script in multiple parts.
        This can be used for client-server interactions, or when chunking a large
        visualization into multiple files. Immediately after this, a
        call to get_script() will returnempty.
        """

    def get_json(self) -> str:
        """Returns a JSON-serialized representation of the Scene as a string.

        If you use clear_script() then this will return updates to
        the Scene since thelast call to clear_script().

        Returns:
            str: a JSON string representing the Scene
        """

    def get_script(self) -> str:
        """Returns a JSONP (https://en.wikipedia.org/wiki/JSONP) script representing the Scene.

        If you use clear_script() then this will return updates to
        the Scene since the last call to clear_script().

        Returns:
            str: a JSONP script representing the Scene
        """

    def save_as_json(self, path: str) -> None:
        """To view the JSON, you will need to separately code up the wrapper html and provide the scenepic.min.js library file.
        Alternatively, use save_as_html() to make a self-contained HTML file.

        Args:
            path (str): the path to the file on disk
        """

    def save_as_script(self, path: str, standalone: bool) -> None:
        """Save the scene as a ScenePic script file (JavaScript JSONP format).

        To view the script, you will need to separately code up the wrapper html.
        Alternatively, use save_as_html() to make a self-contained HTML file.

        Args:
            path (str): the path to the file on disk
            standalone (bool): whether to make the script standalone by including the library
        """

    def save_as_html(self, path: Optional[str] = None, title: Optional[str] = None,
                     head_html: Optional[str] = None, body_html: Optional[str] = None) -> None:
        """Save the scene as a self-contained html file with no dependencies.

        Args:
            path (str): the path to the file on disk
            title (str, optional): the HTML title. Defaults to "ScenePic".
            head_html (str, optional): the raw HTML to place in the HEAD tag. Defaults to None.
            body_html (str, optional): the raw HTML to place in the BODY tag. Defaults to None.
            script_path (str, optional): desired relative path for the script. A value of None
                                            indicates to embed the script in the HTML page.
                                            Defaults to None.          
            library_path (str, optional): desired relative path for the library. A value of None
                                            indicates to embed the library in the HTML page.
                                            Defaults to None.          
        """

    @property
    def script_cleared(self) -> bool:
        """Whether the script has been cleared."""



import numpy as np
from typing import Optional
from .camera import Camera
from .mesh_info import MeshInfo


class VertexBuffer:
    """Class which provides dictionary access to vertex buffer blocks.

    Description:
        The different parts of the vertex buffer can be accessed via a dictionary
        interface:

        - "pos": the vertex positions
        - "norm": the vertex normals
        - "rgb": the vertex colors (if present)
        - "uv": the vertex uvs (if present)
    """

    def __init__(self, values: np.ndarray):
        """Constructor.

        Args:
            values (np.ndarray): The raw vertex buffer values
        """

    @property
    def shape(self) -> tuple:
        """Shape of the entire vertex buffer."""

    def copy(self) -> "VertexBuffer":
        """Returns a copy of the vertex buffer."""

    def __getitem__(self, key: str) -> np.ndarray:
        """Returns a sub-section of the buffer given the key.

        Args:
            key (str): one of ["pos", "norm", "rgb", "uv"]

        Returns:
            np.ndarray: a subsection of the buffer
        """

    def __setitem__(self, key: str, data: np.ndarray):
        """Sets a subsection of the buffer specified by the key.

        Args:
            key (str): one of ["pos", "norm", "rgb", "uv"]
            data (np.ndarray): The new values to place in the buffer
        """


class Mesh:
    """The basic ScenePic mesh class, containing vertex, triangle, and line buffers.

    To allow for compatibility with Numpy, we use row major order, so each
    data point is stored as a row for efficiency. Vertex buffer contains
    float32 elements with structure: x,y,z, and the normal buffer
    contains elements: nx,ny,nz.

    If shared_color is set then per-vertex color [cr,cg,cb,ca] in the
    color buffer is omitted to save memory.

    If texture_id is None then per-vertex texture coordinates [tu, tv] in
    the uv buffer are omitted to save memory.

    Right handed coordinate system used throughout.

    Triangle winding order is counter clockwise.
    """

    def __init__(self, shared_color: Optional[np.ndarray] = None, texture_id: Optional[str] = None):
        """A Mesh created in this way will not be added to the scene.

        Args:
            shared_color (np.ndarray, optional): shared color for all objects in the mesh. Defaults to None.
            texture_id (str, optional): The ID of the Image to use as a texture. Defaults to None.
        """

    def count_vertices(self) -> int:
        """The number of vertices in the mesh."""

    @property
    def center_of_mass(self) -> np.ndarray:
        """The mean of mesh vertex positions."""

    def reverse_triangle_order(self) -> None:
        """Reverses the winding order of all triangles in this mesh.
        Useful when interoping with existing codebases that use opposite convention.
        """

    def apply_transform(self, transform: np.ndarray) -> None:
        """Apply a 3D homogeneous matrix transform (i.e. 4x4 matrix) to all vertices
            (and appropriately to the normals) in the Mesh.

        Args:
            transform (np.ndarray): a 3D homogeneous transform
        """

    def apply_rotation(self, transform: np.ndarray) -> None:
        """Apply a 3D homogeneous matrix rotation (i.e. 4x4 matrix) to all vertices
            (and appropriately to the normals) in the Mesh.

           Separated out from apply_transform() so that a matrix inverse is not required, for speed.

        Args:
            transform (np.ndarray): a 3D homogeneous rotation matrix
        """

    def get_transformed(self, transform: np.ndarray) -> "Mesh":
        """Get a transformed copy of this Mesh.

        Args:
            transform (np.ndarray): a 3D homogeneous transform

        Returns:
            Mesh: a transformed copy of this mesh
        """

    def get_rotated(self, transform: np.ndarray) -> "Mesh":
        """Get a rotated copy of this Mesh.

        Args:
            transform (np.ndarray): a 3D homogeneous rotation matrix

        Returns:
            Mesh: a rotated copy of this mesh
        """

    def append_mesh(self, mesh: "Mesh") -> None:
        """Append another mesh object to this mesh.

        Args:
            mesh (Mesh): the mesh to append
        """

    def add_triangle(self, color: Optional[np.ndarray] = None,
                     p0: np.ndarray = [0, 0, 0], p1: np.ndarray = [1, 0, 0], p2: np.ndarray = [0, 1, 0],
                     normal: Optional[np.ndarray] = None, fill_triangles: bool = True, add_wireframe: bool = False,
                     uv0: np.ndarray = (0, 0), uv1: np.ndarray = (1, 0), uv2: np.ndarray = (0, 1)) -> None:
        """Adds a triangle to the mesh.

        Args:
            color (np.ndarray, optional): required unless Mesh was constructed with shared_color argument. Defaults to None.
            p0 (np.ndarray, optional): the first vertex. Defaults to [0, 0, 0].
            p1 (np.ndarray, optional): the second vertex. Defaults to [1, 0, 0].
            p2 (np.ndarray, optional): the third vertex. Defaults to [0, 1, 0].
            normal (np.ndarray, optional): will be computed automatically if None. Defaults to None.
            fill_triangles (bool, optional): whether to fill the primitive. Defaults to True.
            add_wireframe (bool, optional): whether to add a wireframe outline. Defaults to False.             
            uv_0 (np.ndarray, optional): first UV coordinate (for UV meshes). Defaults to (0, 0).
            uv_1 (np.ndarray, optional): second UV coordinate (for UV meshes). Defaults to (1, 0).
            uv_2 (np.ndarray, optional): third UV coordinate (for UV meshes). Defaults to (0, 1).
        """

    def add_quad(self, color: Optional[np.ndarray] = None,
                 p0: np.ndarray = [0, -0.5, -0.5], p1: np.ndarray = [0, 0.5, -0.5], p2: np.ndarray = [0, 0.5, 0.5], p3: np.ndarray = [0, -0.5, 0.5],
                 normal: Optional[np.ndarray] = None, fill_triangles: bool = True, add_wireframe: bool = False, transform: Optional[np.ndarray] = None,
                 uv_0: np.ndarray = (0, 0), uv_1: np.ndarray = (1, 0), uv_2: np.ndarray = (1, 1), uv_3: np.ndarray = (0, 1)) -> None:
        """Adds a quad to the mesh.
        All points are assumed to lie on the same plane.

        Args:
            color (np.ndarray, optional): required unless Mesh was constructed with shared_color argument. Defaults to None.
            p0 (np.ndarray, optional): the first vertex. Defaults to [0, -0.5, -0.5].
            p1 (np.ndarray, optional): the second vertex. Defaults to [0, 0.5, -0.5].
            p2 (np.ndarray, optional): the third vertex. Defaults to [0, 0.5, 0.5].
            p3 (np.ndarray, optional): the fourth vertex. Defaults to [0, -0.5, 0.5].
            normal (np.ndarray, optional): will be computed automatically if None. Defaults to None.
            fill_triangles (bool, optional): whether to fill the primitive. Defaults to True.
            add_wireframe (bool, optional): whether to add a wireframe outline. Defaults to False.
            transform (np.ndarray, optional): 3D homogeneous transform matrix. Defaults to None.
            uv_0 (np.ndarray, optional): first UV coordinate (for UV meshes). Defaults to (0, 0).
            uv_1 (np.ndarray, optional): second UV coordinate (for UV meshes). Defaults to (1, 0).
            uv_2 (np.ndarray, optional): third UV coordinate (for UV meshes). Defaults to (1, 1).
            uv_3 (np.ndarray, optional): fourth UV coordinate (for UV meshes). Defaults to (0, 1).        
        """

    def add_image(self, origin: np.ndarray = (-0.5, -0.5, 0), x_axis: np.ndarray = (1, 0, 0),
                  y_axis: np.ndarray = (0, 1, 0), normal: Optional[np.ndarray] = None,
                  uv_0: np.ndarray = (0, 0), uv_1: np.ndarray = (1, 0), uv_2: np.ndarray = (1, 1), uv_3: np.ndarray =(0, 1),
                  double_sided: bool = True, transform: Optional[np.ndarray] = None) -> None:
        """Add an image as a textured quad to this Mesh.
        The Mesh must have a texture_id.

        Args:
            origin (np.ndarray, optional): bottom left corner of image. Defaults to (-0.5, -0.5, 0).
            x_axis (np.ndarray, optional): direction and length of the x-axis of the image in world space.
                                            Defaults to (1, 0, 0).
            y_axis (np.ndarray, optional): direction and length of the y-axis of the image in world space.
                                            Defaults to (0, 1, 0).
            normal (np.ndarray, optional): will be computed automatically if None. Defaults to None.
            uv_0 (np.ndarray, optional): the bottom left uv coordinate. Defaults to (0, 0).
            uv_1 (np.ndarray, optional): the top left uv coordinate. Defaults to (1, 0).
            uv_2 (np.ndarray, optional): the tp right uv coordinate. Defaults to (1, 1).
            uv_3 (np.ndarray, optional): the bottom right uv coordinate. Defaults to (0, 1).
            double_sided (bool, optional): whether the image should be shown on both sides. Defaults to True.
            transform (np.ndarray, optional): a 3D homogeneous transform matrix. Defaults to None.
        """

    def add_cube(self, color: Optional[np.ndarray] = None, transform: Optional[np.ndarray] = None,
                 fill_triangles: bool = True, add_wireframe: bool = False) -> None:
        """Add a unit diameter cube to this mesh.

        Args:
            color (np.ndarray, optional): required unless Mesh was constructed with shared_color argument. Defaults to None.
            transform (np.ndarray, optional): homogeneous 3D transform matrix. Defaults to None.
            fill_triangles (bool, optional): whether to fill the primitive. Defaults to True.
            add_wireframe (bool, optional): whether to add a wireframe outline. Defaults to False.
        """

    def add_thickline(self, color: Optional[np.ndarray] = None, start_point: np.ndarray = [-0.5, 0, 0],
                      end_point: np.ndarray = [0.5, 0, 0], start_thickness: float = 0.1,
                      end_thickness: float = 0.1, transform: Optional[np.ndarray] = None,
                      fill_triangles: bool = True, add_wireframe: bool = False) -> None:
        """Add a thick line to this Mesh.

        Args:
            color (np.ndarray, optional): required unless Mesh was constructed with shared_color argument. Defaults to None.
            start_point (np.ndarray, optional): the start point of the line. Defaults to [-0.5, 0, 0].
            end_point (np.ndarray, optional): the end point of the line. Defaults to [0.5, 0, 0].
            start_thickness (float, optional): the start thickness of the line. Defaults to 0.1.
            end_thickness (float, optional): the end thickness of the line. Defaults to 0.1.
            transform (np.ndarray, optional): homogeneous 3D transform matrix. Defaults to None.
            fill_triangles (bool, optional): whether to fill the primitive. Defaults to True.
            add_wireframe (bool, optional): whether to add a wireframe outline. Defaults to False.
        """

    def add_cone(self, color: Optional[np.ndarray] = None, transform: Optional[np.ndarray] = None,
                 truncation_height: float = 1.0, lat_count: int = 10, long_count: int = 10,
                 fill_triangles: bool = True, add_wireframe: bool = False) -> None:
        """Add a cone to this Mesh.
        The cone has base center at [0.5, 0., 0.] and apex at [-0.5, 0., 0.].

        Args:
            color (np.ndarray, optional): required unless Mesh was constructed with shared_color argument. Defaults to None.
            transform (np.ndarray, optional): homogeneous 3D transform. Defaults to None.
            truncation_height (float, optional): draws a truncated cone of this height. Defaults to 1.
            lat_count (int, optional): number of discrete samples in latitude. Defaults to 10.
            long_count (int, optional): number of discrete samples in longitude. Defaults to 10.
            fill_triangles (bool, optional): whether to fill the primitive. Defaults to True.
            add_wireframe (bool, optional): whether to add a wireframe outline. Defaults to False.
        """

    def add_coordinate_axes(self, length: float, thickness: float, transform: Optional[np.ndarray] = None) -> None:
        """Add a set of coordinate axes to this Mesh.
        xyz axes map to RGB colors.

        Args:
            length (float, optional): length of axes. Defaults to 1.0.
            thickness (float, optional): thickness of axes. Defaults to 0.1.
            transform (np.ndarray, optional): homogeneous 3D transform. Defaults to None.
        """

    def add_camera_frustum(self, camera: Camera, color: Optional[np.ndarray] = None, fov_y_degrees: float = 45,
                           aspect_ratio: float = 16/9, depth: float = 1.0, thickness: float = 0.02,
                           transform: Optional[np.ndarray] = None) -> None:
        """Adds a camera frustum visualization to this mesh.

        You can either specify the camera parameters using a Camera object
        or manually via fov_y_degrees, aspect_ratio, and transform.

        Args:
            camera (Camera): the camera to use for the frustum
            color (np.ndarray): required unless Mesh was constructed with shared_color argument. Defaults to None.
            depth (float, optional): depth of the frustum. Defaults to 1.0.
            thickness (float, optional): thickness of the edges. Defaults to 0.02.
            fov_y_degrees (float, optional): vertical field of view in degrees. Defaults to 45.0.
            aspect_ratio (float, optional): aspect ratio of field of view (width/height). Defaults to 16/9.
            transform (np.ndarray, optional): homogeneous 3D transform. Defaults to None.
        """

    def add_camera_image(self, camera: Camera, depth: float) -> None:
        """Add the image plane for a camera into the scene.

        Args:
            camera (Camera, optional): the ScenePic Camera corresponding to the image. 
            depth (float, optional): depth of the frustum. Defaults to 1.0.
        """

    def add_disc(self, color: Optional[np.ndarray] = None, transform: Optional[np.ndarray] = None, segment_count: int = 10,
                 fill_triangles: bool = True, add_wireframe: bool = False) -> None:
        """Add a x-axis aligned unit diameter disc to this Mesh.

        Args:
            color (np.ndarray, optional): required unless Mesh was constructed with shared_color argument. Defaults to None.
            transform (np.ndarray, optional): homogeneous 3D transform. Defaults to None.
            segment_count (int, optional): number of segments in the disc. Defaults to 10.
            fill_triangles (bool, optional): whether to fill the primitive. Defaults to True.
            add_wireframe (bool, optional): whether to add a wireframe outline. Defaults to False.
        """

    def add_cylinder(self, color: Optional[np.ndarray] = None, transform: Optional[np.ndarray] = None, segment_count: int = 10,
                     fill_triangles: bool = True, add_wireframe: bool = False) -> None:
        """Add a x-axis aligned unit diameter cylinder to this Mesh.

        Args:
            color (np.ndarray, optional): required unless Mesh was constructed with shared_color argument. Defaults to None.
            transform (np.ndarray, optional): homogeneous 3D transform. Defaults to None.
            segment_count (int, optional): number of segments in the cylinder. Defaults to 10.
            fill_triangles (bool, optional): whether to fill the primitive. Defaults to True.
            add_wireframe (bool, optional): whether to add a wireframe outline. Defaults to False.
        """

    def add_sphere(self, color: Optional[np.ndarray] = None, transform: Optional[np.ndarray] = None,
                   fill_triangles: bool = True, add_wireframe: bool = False) -> None:
        """Add a default (ico) unit diameter sphere primitive to the mesh.

        Args:
            color (np.ndarray, optional): required unless Mesh was constructed with shared_color argument. Defaults to None.
            transform (np.ndarray, optional): homogeneous 3D transform. Defaults to None.
            fill_triangles (bool, optional): whether to fill the primitive. Defaults to True.
            add_wireframe (bool, optional): whether to add a wireframe outline. Defaults to False.
        """

    def add_icosphere(self, color: Optional[np.ndarray] = None, transform: Optional[np.ndarray] = None, steps: int = 0,
                      fill_triangles: bool = True, add_wireframe: bool= False) -> None:
        """Add a unit diameter ico-sphere to this Mesh.

        Args:
            color (np.ndarray, optional): required unless Mesh was constructed with shared_color argument. Defaults to None.
            transform (np.ndarray, optional): homogeneous 3D transform. Defaults to None.
            steps (int, optional): number of subdivision steps to apply to the icosahedron base primitive. Defaults to 0.
            fill_triangles (bool, optional): whether to fill the primitive. Defaults to True.
            add_wireframe (bool, optional): whether to add a wireframe outline. Defaults to False.
        """

    def add_uv_sphere(self, color: Optional[np.ndarray] = None, transform: Optional[np.ndarray] = None,
                      lat_count: int = 10, long_count: int = 10,
                      fill_triangles: bool = True, add_wireframe: bool = False) -> None:
        """Add a unit diameter UV sphere to this Mesh.

        Description:
            If the mesh has a texture attached, this method will automatically create
            a set of vertex UVs which unwraps the sphere to map to the full UV space.

        Args:
            color (np.ndarray, optional): required unless Mesh was constructed with shared_color argument. Defaults to None.
            transform (np.ndarray, optional): homogeneous 3D transform. Defaults to None.
            lat_count (int, optional): number of discrete samples in latitude. Defaults to 10.
            long_count (int, optional): number of discrete samples in longitude. Defaults to 10.
            fill_triangles (bool, optional): whether to fill the primitive. Defaults to True.
            add_wireframe (bool, optional): whether to add a wireframe outline. Defaults to False.
        """

    def add_mesh(self, mesh_info: MeshInfo, transform: Optional[np.ndarray] = None, reverse_triangle_order: bool = False,
                 fill_triangles: bool = True, add_wireframe: bool = False) -> None:
        """Add a triangle mesh to this ScenePic Mesh, with normals computed automatically.

        Args:
            mesh_info (MeshInfo): object containing the vertices, triangles, and uvs/colors.
            transform (np.ndarray, optional): homogeneous 3D transform. Defaults to None.
            reverse_triangle_order (bool, optional): whether to reverse the triangle winding order. Defaults to False.
            fill_triangles (bool, optional): whether to fill the primitive. Defaults to True.
            add_wireframe (bool, optional): whether to add a wireframe outline. Defaults to False.
        """

    def add_mesh_without_normals(self, vertices: np.ndarray, triangles: np.ndarray,
                                 colors: Optional[np.ndarray] = None, uvs: Optional[np.ndarray] = None,
                                 transform: Optional[np.ndarray] = None, reverse_triangle_order: bool = False,
                                 fill_triangles: bool = True, add_wireframe: bool = False) -> None:
        """Add a triangle mesh to this ScenePic Mesh, with normals computed automatically.

        Args:
            vertices (np.ndarray): float32 matrix of [N, 3] vertex positions
            triangles (np.ndarray): uint32 matrix of [N, 3] triangle vertex indices
            colors (np.ndarray, optional): float32 matrix of [N, 3] per-vertex RGB color vectors. Defaults to None.
            uvs (np.ndarray, optional): float32 matrix of [N, 2] per-vertex UV coordinates. Defaults to None.
            transform (np.ndarray, optional): homogeneous 3D transform. Defaults to None.
            reverse_triangle_order (bool, optional): whether to reverse the triangle winding order. Defaults to False.
            fill_triangles (bool, optional): whether to fill the primitive. Defaults to True.
            add_wireframe (bool, optional): whether to add a wireframe outline. Defaults to False.
        """

    def add_mesh_with_normals(self, vertices: np.ndarray, normals: np.ndarray,
                              triangles: np.ndarray, colors: Optional[np.ndarray] = None, uvs: Optional[np.ndarray] = None,
                              transform: Optional[np.ndarray] = None, reverse_triangle_order: bool = False,
                              fill_triangles: bool = True, add_wireframe: bool = False) -> None:
        """Add a triangle mesh to this ScenePic Mesh.

        Args:
            vertices (np.ndarray): float32 matrix of [N, 3] vertex positions
            normals (np.ndarray): float32 matrix of [N, 3] vertex normals
            triangles (np.ndarray): uint32 matrix of [N, 3] triangle vertex indices
            colors (np.ndarray, optional): float32 matrix of [N, 3] per-vertex RGB color vectors. Defaults to None.
            uvs (np.ndarray, optional): float32 matrix of [N, 2] per-vertex UV coordinates. Defaults to None.
            transform (np.ndarray, optional): homogeneous 3D transform. Defaults to None.
            reverse_triangle_order (bool, optional): whether to reverse the triangle winding order. Defaults to False.
            fill_triangles (bool, optional): whether to fill the primitive. Defaults to True.
            add_wireframe (bool, optional): whether to add a wireframe outline. Defaults to False.
        """

    def add_lines(self, start_points: np.ndarray, end_points: np.ndarray,
                  color: Optional[np.ndarray] = None, transform: Optional[np.ndarray] = None) -> None:
        """Add a line cloud to this Mesh.

        Args:
            start_points (np.ndarray): float32 matrix of [N, 3] line start positions (or [N, 6] positions with colors)
            end_points (np.ndarray): float32 matrix of [N, 3] line end positions (or [N, 6] positions with colors)
            color (np.ndarray, optional): if provided, will override any per-point colors stored in start_colors and end_colors.
                                            Defaults to None.
            transform (np.ndarray): homogeneous 3D transform. Defaults to None.
        """

    def enable_instancing(self, positions: np.ndarray, rotations: Optional[np.ndarray] = None,
                          colors: Optional[np.ndarray] = None) -> None:
        """Makes ScenePic render this Mesh with multiple instances, e.g. for point-cloud visualizations.
        Can be used to make point clouds, for example.
        The whole contents of the Mesh will be rendered multiple times.

        Args:
            positions (np.ndarray): float32 matrix of [N, 3] instance positions
            rotations (np.ndarray, optional): float32 matrix of [N, 4] per-point quaternion rotations. Defaults to None.
            colors (np.ndarray, optional): float32 matrix of [N, 3] colors. If provided, overrides a Mesh's shared_color
                                            or per-vertex color. Defaults to None.
        """

    @property
    def shared_color(self) -> np.ndarray:
        """Whole-mesh color (reduces memory requirements but makes Mesh monochrome)."""

    @property
    def texture_id(self) -> str:
        """Specifies the id of a ScenePic Image to be used as a texture in this Mesh."""

    @property
    def mesh_id(self) -> str:
        """A unique identifier for the Mesh"""

    @property
    def layer_id(self) -> str:
        """An identifier to allow several Meshes to be grouped together into a "layer".

        Layers should be used in conjunction with Canvas3D.set_layer_settings().
        """

    @property
    def double_sided(self) -> bool:
        """Whether to turn off back face culling and draw the Mesh's triangles as double sided."""

    @property
    def camera_space(self) -> bool:
        """Whether this Mesh is defined in camera space (cannot be moved in the ScenePic user interface) or world space (standard)."""

    @property
    def nn_texture(self) -> bool:
        """ Whether the shader should use nearest-neighbor texture interpolation."""

    @property
    def use_texture_alpha(self) -> bool:
        """Whether to use the alpha channel in the texture for transparency (only relevant for textured Meshes)."""

    @property
    def is_billbord(self) -> bool:
        """Draw this Mesh as a billboard (i.e. always facing the user) rather than rotating with the rest of the world."""

    @property
    def vertex_buffer(self) -> VertexBuffer:
        """The raw vertex buffer."""


class MeshUpdate:
    """A new mesh formed by updated an existing mesh."""

    @property
    def base_mesh_id(self) -> str:
        """The unique identifier for the base mesh"""

    @property
    def mesh_id(self) -> str:
        """The unique identifier of the newly created mesh."""

    @property
    def frame_index(self) -> int:
        """The position in the creation order for this update."""

    @property
    def vertex_buffer(self) -> VertexBuffer:
        """The raw vertex buffer."""

    def quantize(self, keyframe_index: int, fixed_point_range: float, keyframe_vertex_buffer: VertexBuffer):
        """Quantize the mesh update.

        Args:
            self (MeshUpdate): self reference
            keyframe_index (int): Index of the keyframe to use in quantizing this update
            fixed_point_range (float): The range to use for the fixed point representation.
            keyframe_vertex_buffer (VertexBuffer): The keyframe vertex buffer
        """

    def difference_range(self, vertex_buffer: VertexBuffer) -> float:
        """Returns the absolute range of values in the difference between this update and the buffer.

        Args:
            self (MeshUpdate): self reference
            vertex_buffer (VertexBuffer): the buffer to use in the comparison

        Return:
            float: the absolute range (from minimum to maximum) in the per-index difference between
                this update and the reference.
        """



from typing import List

import numpy as np


class Graph:
    """A 2D viewport that animates one or more sparklines."""

    class Margin:
        """Represents the margin along the edges of a graph."""

        def __init__(self, size: float, top: float, right: float,
                     bottom: float, left: float):
            """Constructor.

            Args:
                size (float, optional): space in pixels to use for all margins. Default 10.
                top (float, optional): the top margin value in pixels
                right (float, optional): the right margin value in pixels
                bottom (float, optional): the bottom margin value in pixels
                left (float, optional): the left margin value in pixels
            """

        top: float
        """The top margin in pixels."""
        right: float
        """The top margin in pixels."""
        bottom: float
        """The top margin in pixels."""
        left: float
        """The top margin in pixels."""

    class VerticalRule:
        """Represents a vertical line in a sparkline graph."""

        def __init__(self, frame: int, color: np.ndarray, line_width: float):
            """Constructor.

            Args:
                frame (int): Frame to add the vertical rule
                color: (Color, optional): the line color. Defauts to black.
                line_width: (float, optional): The line width in pixels. Defaults to 1.0f.
            """

        frame: int
        """The frame at which to add the line."""
        color: np.ndarray
        """The color of the line."""
        line_width: float
        """The width of the line in pixels."""

    @property
    def canvas_id(self) -> str:
        """A unique identifier for the canvas."""

    @property
    def background_color(self) -> np.ndarray:
        """The background color of the frame."""

    @property
    def margin(self) -> Margin:
        """The outside margin of the graph."""

    @property
    def font_family(self) -> str:
        """The font family used for the graph labels."""

    @property
    def name_size(self) -> float:
        """The size of the sparkline labels in pixels."""

    @property
    def value_size(self) -> float:
        """The size of the sparkline values in pixels."""

    @property
    def name_align(self) -> str:
        """The alignment of the sparkline labels.
        One of ('left', 'right', 'top', 'bottom', or 'none')
        """

    @property
    def value_align(self) -> str:
        """The alignment of the sparkline values.        
        One of ('left', 'right', 'top', 'bottom', or 'none')
        """


    @property
    def media_id(self) -> str:
        """The ID of the attached media file.

        This file will be used to drive playback, i.e. frames will be
        displayed in time with the playback of the media file.
        """

    def add_sparkline(self, name: str, values: np.ndarray,
                      line_color: np.ndarray, line_width: float,
                      vertical_rules: List[VerticalRule] = None) -> None:
        """Adds a sparkline to the graph.

        Args:
            name (str): the name of the measured quantity
            values (np.ndarray): the measured values
            line_color (Color, optional): the color of the line (and its labels). Defaults to Black.
            line_width (float, optional): The width of the line. Defaults to 1.0.
            vertical_rules (List[VerticalRule], optional): the vertical rules to add. Defaults to None.
        """



import numpy as np


class FocusPoint:
    """3D focus point for this frame (with optional 3D rotation for orientation lock), used in the ScenePic user interface."""

    def __init__(self, positions: np.ndarray, orientation_axis_angle: np.ndarray):
        """Constructor.

        Args:
            positions (np.ndarray): The focus point position
            orientation_axis_angle (np.ndarray, optional): Orientation axis for orientation lock
        """

    @property
    def position(self) -> np.array:
        """The position of the focus point."""

    @property
    def orientation_axis_angle(self) -> np.ndarray:
        """The orientation axis angle"""



from typing import List, Mapping, Union

import numpy as np
from typing import Optional

from .camera import Camera
from .focus_point import FocusPoint
from .label import Label
from .mesh import Mesh, MeshUpdate
from .layer_settings import LayerSettings


class Frame3D:
    """Represents a frame of an animation which contains a number of transformed Mesh objects."""

    def add_meshes(self, meshes: List[Union[Mesh, MeshUpdate, str]], transform: Optional[np.ndarray] = None):
        """Adds meshes to the frame.

        Args:
            meshes (List[Union[Mesh, MeshUpdate, str]]): A list of meshes and mesh updates (or their IDs)
            transform (np.ndarray, optional): an optional transform to apply to each mesh. Defaults to None.
        """

    def add_mesh(self, mesh: Union[Mesh, MeshUpdate, str], transform: Optional[np.ndarray] = None):
        """Adds a mesh to the frame.

        Args:
            mesh (Union[Mesh, MeshUpdate, str]): The mesh, mesh update, or mesh ID
            transform (np.ndarray, optional): An optional transform to apply to the mesh. Defaults to None.
        """

    def add_label(self, label: Label, position: np.ndarray) -> None:
        """Add a label to the frame.

        Args:
            label (Label): the label to add
            position (np.ndarray): the float32 [x, y, z] position at which to put the label
        """

    def set_layer_settings(self, layer_settings: Mapping[str, Union[dict, LayerSettings]]):
        """Specify the visibilities and opacities of certain mesh layers.

        Description:
            Each Mesh object can optionally be part of a user-identified layer
            (see Mesh constructor).
            Calling set_layer_settings will result in an additional drop-down
            selector in the ScenePic user interface.

        Args:
            layer_settings (Mapping[str, Union[dict, LayerSettings]]): a LayerSettings object
                or a dictionary. See LayerSettings for details.
        """

    @property
    def focus_point(self) -> FocusPoint:
        """The focus point of the frame"""

    @property
    def camera(self) -> Camera:
        """The camera for the frame.

        This property can be used to create cinematic camera movement within a
        ScenePic, but the user can choose to override the camera
        track at any time. If they reset to the original camera, however, it will make
        subsequent frames use the specified camera parameters.
        """



from typing import Union

import numpy as np
from typing import Optional

from .image import Image


class Frame2D:
    """Represents a frame of a 2D animation."""

    def add_line(self, coordinates: np.ndarray, line_color: np.ndarray,
                 line_width: float, fill_color: Optional[np.ndarray] = None, close_path: bool = False ,
                 layer_id: str = "") -> None:
        """Add a line to the frame.

        Args:
            coordinates (np.ndarray): float32 matrix of [N, 2] coordinates of the line
            line_color (np.ndarray, optional): the color of the line. Defaults to Black.
            line_width (float, optional): the width of the line. Defaults to 1.0.
            fill_color (np.ndarray, optional): the fill color to use. Defaults to None.
            close_path (bool, optional): whether to close the path. Defaults to False.
            layer_id (str, optional): Unique ID of the layer associated with this primitive. Defaults to "".
        """

    def add_rectangle(self, x: float, y: float, w: float, h: float,
                      line_color: np.ndarray, line_width: float,
                      fill_color: Optional[np.ndarray] = None, layer_id: str = "") -> None:
        """Add a rectangle to the frame.

        Args:
            x (float): the upper left corner x coordinate
            y (float): the upper left corner y coordinate
            w (float): the width of the rectangle
            h (float): the height of the rectangle
            line_color (np.ndarray, optional): the color of the line. Defaults to Black.
            line_width (float, optional): the width of the line. Defaults to 1.0.
            fill_color (np.ndarray, optional): the fill color to use. Defaults to None.
            layer_id (str, optional): Unique ID of the layer associated with this primitive. Defaults to "".
        """

    def add_video(self, position_type: str, x: float, y: float, scale: float,
                  smoothed: bool, layer_id: str) -> None:
        """Add an frame from a video to the frame.

        Args:
            position_type (str, optional): one of "fit", "fill", "stretch", or "manual". Defaults to "fit".
            x (float, optional): x coordinate to draw video at if position_type == "manual". Defaults to 0.0.
            y (float, optional): y coordinate to draw video at if position_type == "manual". Defaults to 0.0.
            scale (float, optional): scale factor to draw video by if position_type == "manual". Defaults to 1.0.
            smoothed (bool, optional): whether to draw an video with smoothing or using nearest neighbor. Defaults to False.
            layer_id (str, optional): Unique ID of the layer associated with this primitive. Defaults to "".
        """

    def add_image(self, image_id: Union[Image, str], position_type: str,
                  x: float, y: float, scale: float, smoothed: bool,
                  layer_id: str) -> None:
        """Add an image to the frame.

        Args:
            self (Frame2D): the self reference
            image_id (str): the unique identifier of the image
            position_type (str, optional): one of "fit", "fill", "stretch", or "manual". Defaults to "fit".
            x (float, optional): x coordinate to draw image at if position_type == "manual". Defaults to 0.0.
            y (float, optional): y coordinate to draw image at if position_type == "manual". Defaults to 0.0.
            scale (float, optional): scale factor to draw image by if position_type == "manual". Defaults to 1.0.
            smoothed (bool, optional): whether to draw an image with smoothing or using nearest neighbor. Defaults to False.
            layer_id (str, optional): Unique ID of the layer associated with this primitive. Defaults to "".
        """

    def add_circle(self, x: float, y: float, radius: float,
                   line_color: np.ndarray, line_width: float,
                   fill_color: Optional[np.ndarray] = None, layer_id: str = "") -> None:
        """Add a circle to the frame.

        Args:
            self (Frame2D): the self reference
            x (float): the center x coordinate
            y (float): the center y coordinate
            radius (float): the radius of the circle
            line_color (np.ndarray, optional): the color of the line. Defaults to Black.
            line_width (float, optional): the width of the line. Defaults to 1.0.
            fill_color (np.ndarray, optional): the fill color to use. Defaults to None.
            layer_id (str, optional): Unique ID of the layer associated with this primitive. Defaults to "".
        """

    def add_text(self, text: str, left: float, bottom: float,
                 color: np.ndarray, size_in_pixels: float, font_family: str,
                 layer_id: str):
        """Add some text to the frame.

        Args:
            self (Frame2D): The self reference
            text (str): the text to display
            left (float): the pixel position of the left side of the text
            bottom (float): the pixel position of the bottom of the text
            color (np.ndarray, optional): the color of the text. Defaults to White.
            size_in_pixels (float, optional): the vertical size of the text in pixels. Defaults to 12.0.
            font_family (str, optional): the font to use for the text. Defaults to "sans-serif".
            layer_id (str, optional): Unique ID of the layer associated with this primitive. Defaults to "".
        """



import numpy as np


class MeshInfo:
    """Holds information needed to construct a mesh object.
    Includes information like the vertex positions, the triangulation,
    and either per-vertex UV coordinates or Color (if present).
    """

    @property
    def positions(self) -> np.ndarray:
        """The position buffer."""

    @property
    def normals(self) -> np.ndarray:
        """The normal buffer."""

    @property
    def triangles(self) -> np.ndarray:
        """The triangle buffer."""

    @property
    def uvs(self) -> np.ndarray:
        """The UV buffer."""

    @property
    def colors(self) -> np.ndarray:
        """The color buffer."""

    @property
    def has_normals(self) -> bool:
        """Whether the mesh info contains vertex normals."""

    @property
    def subdivide(self, steps: int, project_to_limit: bool) -> "MeshInfo":
        """Subdivide this mesh using loop subdivision.

        Args:
            steps (int): specifies how many steps of subdivision to apply. Defaults to 1.
            project_to_limit (bool): specifies whether the vertices should be projected onto the
                                     limit surface in the final step of subdivision. Defaults to false.

        Returns:
            MeshInfo: a subdivided version of this mesh
        """



import numpy as np


class Label:
    """A ScenePic label type"""

    @property
    def label_id(self) -> str:
        """A unique identifier for the Label"""

    @property
    def mesh_id(self) -> str:
        """The mesh associated with this label."""

    @property
    def text(self) -> str:
        """The text to use in the label."""

    @property
    def fill_color(self) -> np.ndarray:
        """The color to use."""

    @property
    def size_in_pixels(self) -> float:
        """The size in pixels at which to display the label."""

    @property
    def font_family(self) -> str:
        """HTML font family name."""

    @property
    def horizontal_align(self) -> str:
        """How to align the label horizontally (one of 'left', 'center', 'right')"""

    @property
    def offset_distance(self) -> float:
        """Distance in world space to offset the label by when using left/right/top/bottom alignment."""
    


"""Module which extends the scenepic Mesh type."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import numpy as np

from ._scenepic import Mesh, MeshUpdate


class VertexBuffer:
    def __init__(self, values: np.ndarray):
        self._values = values
        self._lookup = {
            "pos": slice(0, 3),
            "norm": slice(3, 6),
            "rgb": slice(6, 9),
            "uv": slice(6, 8)
        }

    @property
    def shape(self) -> tuple:
        return self._values.shape

    def copy(self) -> "VertexBuffer":
        return VertexBuffer(self._values.copy())

    def __repr__(self) -> str:
        return str(self._values)

    def __getitem__(self, key: str) -> np.ndarray:
        assert key in self._lookup, "Invalid vertex buffer key " + key
        return self._values[:, self._lookup[key]]

    def __setitem__(self, key: str, data: np.ndarray):
        assert key in self._lookup, "Invalid vertex buffer key " + key
        self._values[:, self._lookup[key]] = data


def vertex_buffer(self):
    return VertexBuffer(self.get_vertex_buffer())


def quantize(self, keyframe_index: int, fixed_point_range: float, keyframe_vertex_buffer: VertexBuffer):
    self.quantize_(keyframe_index, fixed_point_range, keyframe_vertex_buffer._values)


def difference_range(self, vertex_buffer: VertexBuffer) -> float:
    self.difference_range_(vertex_buffer._values)


Mesh.vertex_buffer = property(vertex_buffer, doc="""The raw vertex buffer.""")

MeshUpdate.vertex_buffer = property(vertex_buffer, doc="""The raw vertex buffer.""")

MeshUpdate.quantize = quantize
MeshUpdate.quantize.__doc__ = """Quantize the mesh update.

Args:
    self (MeshUpdate): self reference
    keyframe_index (int): Index of the keyframe to use in quantizing this update
    fixed_point_range (float): The range to use for the fixed point representation.
    keyframe_vertex_buffer (VertexBuffer): The keyframe vertex buffer
"""

MeshUpdate.difference_range = difference_range
MeshUpdate.difference_range.__doc__ = """Returns the absolute range of values in the difference between this update and the buffer.

Args:
    self (MeshUpdate): self reference
    vertex_buffer (VertexBuffer): the buffer to use in the comparison

Return:
    float: the absolute range (from minimum to maximum) in the per-index difference between
        this update and the reference.
"""



from typing import Mapping, Union

import numpy as np

from .frame2d import Frame2D
from .layer_settings import LayerSettings


class Canvas2D:
    """A 2D canvas that can contain a number of Frame2Ds."""

    @property
    def canvas_id(self) -> str:
        """A unique identifier for the canvas."""

    def create_frame(self, frame_id: str) -> Frame2D:
        """Creates a new Frame2D object and appends to the list of Frame2Ds in the Canvas2D.

        Args:
            frame_id (str, optional): a unique identifier for the Frame2D (will be
                                      automatically populated if not provided).

        Returns:
            Frame2D: a new Frame2D object
        """

    def set_layer_settings(self, layer_settings: Mapping[str, Union[dict, LayerSettings]]):
        """Specify the visibilities and opacities of certain mesh layers.

        Description:
            Each Mesh object can optionally be part of a user-identified layer
            (see Mesh constructor).
            Calling set_layer_settings will result in an additional drop-down
            selector in the ScenePic user interface.

        Args:
            self (Canvas2D): the self reference
            layer_settings (Mapping[str, Union[dict, LayerSettings]]): a mapping from
                layer IDs to layer settings.
        """

    def clear_script(self) -> None:
        """Call this if you are dumping a ScenePic script in multiple parts.

        Use this for client-server interactions, or when chunking a large
        visualization into multiple files.
        """

    @property
    def width(self) -> float:
        """The width of the Canvas."""

    @property
    def height(self) -> float:
        """The height of the Canvas."""

    @property
    def background_color(self) -> np.ndarray:
        """The background color of the frame."""

    @property
    def media_id(self) -> np.ndarray:
        """The ID of the attached media file.

        This file will be used to drive playback, i.e. frames will be
        displayed in time with the playback of the media file.
        """



"""Module which extends the scenepic AudioTrack type."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from ._scenepic import AudioTrack


def load_from_buffer(self, data: bytes, ext: str):
    self.data = list(data)
    self.ext = ext


def _data_get(self):
    return bytes(self.data_get_)


def _data_set(self, data):
    self.data_set_(list(data))


AudioTrack.load_from_buffer = load_from_buffer
AudioTrack.load_from_buffer.__doc__ = """Load audio from a pre-existing data buffer and extension.

Args:
    self (AudioTrack): self reference
    data (bytes): the binary data for the audio
    ext (str): the file extension (e.g. mp3, ogg, wav)
"""

AudioTrack.data = property(_data_get, _data_set, doc="""The binary audio data.""")



class Video:
    """A ScenePic Video type."""

    def load(self, path: str) -> None:
        """Load a video file from the disk.

        Args:
            path (str): the path to the video file
        """

    def load_from_buffer(self, data: bytes, ext: str) -> None:
        """Load video from a pre-existing data buffer and extension.

        Args:
            data (bytes): the binary data for the video
            ext (str): the file extension (e.g. mp4, mkv)
        """

    @property
    def audio_id(self) -> str:
        """A unique identifier for the video."""

    @property
    def data(self) -> bytes:
        """The binary video data."""



from typing import List, Mapping, Union

from .focus_point import FocusPoint
from .frame3d import Frame3D
from .camera import Camera
from .layer_settings import LayerSettings
from .mesh import Mesh, MeshUpdate
from .shading import Shading
from .ui_parameters import UIParameters


class Canvas3D:
    """Represents a ScenePic Canvas3D - a 3D viewport that can contain a number of Frame3Ds."""

    def create_frame(self, frame_id: str, focus_point: FocusPoint,
                     meshes: List[Union[Mesh, MeshUpdate, str]],
                     camera: Camera) -> Frame3D:
        """Creates a new Frame3D object and appends to the list of Frame3Ds in the Canvas3D.

        Args:
            frame_id (str): a unique identifier for the Frame3D (will be
                            automatically populated if not provided).
            focus_point (FocusPoint): optional 3D focus point for this frame (with optional 3D
                                    rotation for orientation lock), used in the ScenePic user interface.
            meshes (List[Union[Mesh, MeshUpdate, str]]): optionally pre-populate this Frame3D
                                                        with a set of Meshes, for convenience.
            camera (Camera): optionally set a camera for this frame

        Returns:
            Frame3D: a new Frame3D object
        """

    def set_layer_settings(self, layer_settings: Mapping[str, Union[dict, LayerSettings]]):
        """Specify the visibilities and opacities of certain mesh layers.

        Description:
            Each Mesh object can optionally be part of a user-identified layer
            (see Mesh constructor).
            Calling set_layer_settings will result in an additional drop-down
            selector in the ScenePic user interface.

        Args:
            self (Canvas3D): the self reference
            layer_settings (Mapping[str, Union[dict, LayerSettings]]): a LayerSettings object
                or a dictionary. See LayerSettings for details.
        """

    @property
    def canvas_id(self) -> str:
        """A unique identifier for the canvas."""

    def clear_script(self) -> None:
        """Call this if you are dumping a ScenePic script in multiple parts.

        Use this for client-server interactions, or when chunking a large
        visualization into multiple files.
        """

    @property
    def width(self) -> float:
        """The width of the Canvas."""

    @property
    def height(self) -> float:
        """The height of the Canvas."""

    @property
    def camera(self) -> Camera:
        """Set of Camera parameters for the Canvas."""

    @property
    def shading(self) -> Shading:
        """Set of Shading parameters for the Canvas."""

    @property
    def focus_point(self) -> FocusPoint:
        """Default focus point for the canvas."""

    @property
    def ui_parameters(self) -> UIParameters:
        """Set of user interface parameters."""

    @property
    def media_id(self) -> str:
        """The ID of the attached media file.

        This file will be used to drive playback, i.e. frames will be
        displayed in time with the playback of the media file.
        """



"""Module which extends the scenepic Canvas3D type."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from typing import List, Union, Mapping

import numpy as np

from ._scenepic import (
    Camera,
    Canvas3D,
    FocusPoint,
    Frame3D,
    LayerSettings,
    Mesh,
    MeshUpdate
)


def create_frame(self,
                 frame_id="",
                 focus_point: FocusPoint = None,
                 meshes: List[Union[Mesh, MeshUpdate, str]] = None,
                 camera: Camera = None) -> Frame3D:
    kwargs = {}
    if focus_point is not None:
        if isinstance(focus_point, (list, np.ndarray)):
            focus_point = FocusPoint(focus_point)

        kwargs["focus_point"] = focus_point

    if meshes is not None:
        kwargs["mesh_ids"] = [mesh if isinstance(mesh, str) else mesh.mesh_id
                              for mesh in meshes]

    if camera is not None:
        kwargs["camera"] = camera

    return self.create_frame_(frame_id, **kwargs)


def set_layer_settings(self, layer_settings: Mapping[str, Union[dict, LayerSettings]]):
    updated_settings = {}
    for key, value in layer_settings.items():
        if isinstance(value, dict):
            if "visible" in value:
                print("`visible` keyword has been deprecated")
            else:
                updated_settings[key] = LayerSettings(**value)

    self.set_layer_settings_(updated_settings)


Canvas3D.create_frame = create_frame
Canvas3D.create_frame.__doc__ = """Creates a new Frame3D object and appends to the list of Frame3Ds in the Canvas3D.

Args:
    frame_id (str): a unique identifier for the Frame3D (will be
                    automatically populated if not provided).
    focus_point (FocusPoint): optional 3D focus point for this frame (with optional 3D
                            rotation for orientation lock), used in the ScenePic user interface.
    meshes (List[Union[Mesh, MeshUpdate, str]]): optionally pre-populate this Frame3D
                                                with a set of Meshes, for convenience.
    camera (Camera): optionally set a camera for this frame

Returns:
    Frame3D: a new Frame3D object
"""

Canvas3D.set_layer_settings = set_layer_settings
Canvas3D.set_layer_settings.__doc__ = """Specify the visibilities and opacities of certain mesh layers.

Description:
    Each Mesh object can optionally be part of a user-identified layer
    (see Mesh constructor).
    Calling set_layer_settings will result in an additional drop-down
    selector in the ScenePic user interface.

Args:
    self (Canvas3D): the self reference
    layer_settings (Mapping[str, Union[dict, LayerSettings]]): a LayerSettings object
        or a dictionary. See LayerSettings for details.
"""



from typing import List

import numpy as np


class Camera:
    """Represents a camera in the scene."""

    def __init__(self, center: np.ndarray, look_at: np.ndarray, up_dir: np.ndarray,
                 fov_y_degrees: float, near_crop_distance: float,
                 far_crop_distance: float, rotation: np.ndarray,
                 world_to_camera: np.ndarray, projection: np.ndarray,
                 aspect_ratio: float):
        """Constructor.

        Note:
            The camera can be created either with the look at paradigm (center,
            look_at, up_dir), with center and rotation, or with an explicit
            world-to-camera matrix.

        Args:
            center (np.ndarray, optional): the camera center. Defaults to [0, 0, 4].
            look_at (np.ndarray, optional): where the camera is looking. Defaults to [0, 0, 0].
            up_dir (np.ndarray, optional): the direction that is "up" for the camera. Defaults to [0, 1, 0].
            fov_y_degrees (float, optional): vertical field of view in degrees. Defaults to 45.0.
            near_crop_distance (float, optional): the near clipping plane of the view frustum. Defaults to 0.01.
            far_crop_distance (float, optional): the far clipping plane of the view frustum. Defaults to 20.0.
            rotation (np.ndarray, optional): a homogeneous rotation transform.
            world_to_camera (np.ndarray, optional): a homogeneous 3D transform mapping world coordinates to camera space.
            projection (np.ndarray, optional): the WebGL projection matrix from camera view coordinates to NDC.
            aspect_ratio (float, optional): The aspect ratio of the camera. Defaults to 1.0.
        """

    @property
    def aspect_ratio(self) -> float:
        """The aspect ratio of the camera."""

    @property
    def center(self) -> np.ndarray:
        """The camera center."""

    @property
    def look_at(self) -> np.ndarray:
        """Where the camera is looking."""

    @property
    def up_dir(self) -> np.ndarray:
        """The direction that is \"up\" for the camera."""

    @property
    def rotation(self) -> np.ndarray:
        """The rotation matrix."""

    @property
    def world_to_camera(self) -> np.ndarray:
        """Homogeneous 3D transform mapping world coordinates to camera space."""

    @property
    def camera_to_world(self) -> np.ndarray:
        """Homogeneous 3D transform that places the camera in the scene."""

    @property
    def projection(self) -> np.ndarray:
        """WebGL projection matrix from camera view coordinates to NDC."""

    @staticmethod
    def camera_orbit(num_frames: int, distance: float, num_times: int,
                     min_altitutde: float, max_altitude: float,
                     up_dir: np.ndarray, fowrard_dir: np.ndarray,
                     fov_y_degrees: float, aspect_ratio: float,
                     near_crop_distance: float,
                     far_crop_distance: float) -> List["Camera"]:
        """Creates a sequence of cameras which move in an orbit around the origin.

        Args:
            num_frames (int): the number of frames in the sequence
            distance (float): the distance of the camera from the origin
            num_times (int): the number of times to orbit
            min_altitude (float): the minimum altitude of the orbit
            max_altitude (float): the maximum altitude of the orbit
            up_dir (np.ndarray): the direct that is up (as a normalized vector)
            forward_dir (np.ndarray): the direction that is forward (as a normalized vector)
            fov_y_degrees (float): the field of the view of the camera in degrees
            aspect_ratio (float): the aspect ratio of the camera
            near_crop_distance (float): the near clipping plane of the view frustum
            far_crop_distance (float): the far clipping plane of the view frustum

        Returns:
            List[Camera]: a sequence of cameras moving in an orbit
        """



import numpy as np


class Image:
    """A ScenePic Image type"""

    def load(self, path: str) -> None:
        """Load an image file from the disk

        Args:
            path (str): the path to the image file
        """

    def load_from_buffer(self, data: bytes, ext: str):
        """Load an image from a pre-existing data buffer and extension.

        Args:
            self (Image): self reference
            data (bytes): the binary data for the image
            ext (str): the format of the image data [e.g. "gif", "jpg", "png"]
        """

    def from_numpy(self, array: np.ndarray, format_str: str):  # noqa: A002
        """Loads the image object from raw RGB values provided as a numpy array.

        Args:
            self (Image): self reference
            array (np.ndarray): The raw pixel RGB as either 0-255 bytes or 0-1 floats.
            format_str (str, optional): The format to use when compressing the image. Defaults to "png".
        """

    @property
    def image_id(self) -> str:
        """A unique identifier for the image"""

    @property
    def ext(self) -> str:
        """The extension of the image (e.g. JPG, PNG)"""

    @property
    def data(self) -> bytes:
        """The binary image data."""



import numpy as np

from .mesh_info import MeshInfo


def js_lib_src() -> str:
    """Returns the Javascript ScenePic library as a string."""


def ColorFromBytes(red: int, green: int, blue: int) -> np.ndarray:
    """Construct a color object from byte values.

    Args:
        red (int): Red channel
        green (int): Green channel
        blue (int): Blue channel

    Returns:
        np.ndarray: a valid color object
    """


def load_obj(path: str) -> MeshInfo:
    """Loads a WaveFront OBJ file from disk as a MeshInfo object.

    Args:
        path (str): the path to the OBJ file on disk

    Returns:
        MeshInfo: contains the positions, a triangulation, and UVs (if present)
    """



class LayerSettings:
    """Settings used for customizing canvas drawing by layer.
    
    Currently, opacity is only guaranteed to be correct for Meshes
    that do not overlap.
    If you know one mesh should be rendered before another for opacity purposes,
    use the renderOrder setting: low numbers will be drawn before high numbers,
    and layers with no renderOrder set will be drawn last. Meshes without layer ids,
    or with layer ids that are not specified in this dictionary, will be drawn filled,
    without wireframe, at opacity 1.0, and after any layers with renderOrder set.
    """

    def __init__(self, filled: bool, wireframe: bool,
                 opacity: float, render_order: int):
        """Constructor.

        Args:
            filled (bool, optional): whether to draw the layer with filled polygons. Defaults to True.
            wireframe (bool, optional): whether to draw the layer with a frame outline. Defaults to False.
            opacity (float, optional): the opacity with which to draw the layer. Defaults to 1.0.
            render_order (int, optional): Where in the transparency order to draw this layer. Defaults to None.
        """

    @property
    def filled(self) -> bool:
        """Specifies whether to draw the layer with filled triangles."""

    @property
    def wireframe(self) -> bool:
        """Specifies whether to draw the layer with a wireframe outline."""

    @property
    def opacity(self) -> float:
        """In [0,1] specifies the opacity with which to draw the layer."""

    @property
    def render_order(self) -> int:
        """Specifies a fixed order to draw transparency."""



"""Module which extends the ScenePic Scene type."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import os
from typing import Union

from ._scenepic import (
    Canvas2D,
    Canvas3D,
    Graph,
    Image,
    js_lib_src,
    Mesh,
    Scene
)


def link_canvas_events(self, *canvases):
    ids = [canvas.canvas_id if isinstance(canvas, (Canvas3D, Canvas2D, Graph)) else canvas
           for canvas in canvases]
    self.link_canvas_events_(ids)


def create_mesh(self, mesh_id="", **kwargs) -> Mesh:
    if "texture_id" in kwargs:
        if isinstance(kwargs["texture_id"], Image):
            kwargs["texture_id"] = kwargs["texture_id"].image_id

    return self.create_mesh_(mesh_id, **kwargs)


def simulate_key_presses(self, canvas_id: Union[Canvas3D, Canvas2D, str], *keypresses):
    if isinstance(canvas_id, (Canvas2D, Canvas3D)):
        canvas_id = canvas_id.canvas_id

    self.simulate_key_presses_(canvas_id, keypresses)


def _repr_javascript_(self):
    if self.script_cleared:
        raise Exception("You should not call clear_script() on Scenes that "
                        "you wish to use in Jupyter.")

    # Jupyter notebooks need to load the library as an external file
    with open("scenepic.min.js", "wb") as file:
        src = js_lib_src()
        file.write(src.encode())

    loader_path = os.path.join(os.path.dirname(__file__),
                               "scene_jupyter_template.js")
    with open(loader_path) as file:
        script = file.read() % self.get_json()

    return script


Scene.link_canvas_events = link_canvas_events
Scene.link_canvas_events.__doc__ = """Specify that all input events should be linked across the provided set of Canvases.

Args:
    self (Scene): self reference
    canvases: the canvases which should have linked events
"""

Scene.create_mesh = create_mesh
Scene.create_mesh.__doc__ = """Create a new Mesh and append to the Scene.

Description:
    Meshes are Scene-wide resources that can be reused across multiple
    Frames in multiple Canvas3Ds.

Args:
    mesh_id (str, optional): a unique identifier for the Mesh (will be
                            automatically populated if not provided). Defaults to None.
    layer_id (str, optional): an identifier to allow several Meshes to be grouped together
                            into a "layer".  Layers should be used in conjunction with
                            Canvas3D.set_layer_settings(). Defaults to None.
    double_sided (bool, optional): whether to turn off back face culling and draw the Mesh's
                                triangles as double sided. Defaults to False.
    camera_space (bool, optional): whether this Mesh is defined in camera space (cannot
                                be moved in the ScenePic user interface) or world space (standard).
                                Defaults to False.
    shared_color (np.ndarray, optional): see `Mesh` constructor.
    texture_id (Union[Image, str], optional): see `Mesh` constructor.
    nn_texture (bool, optional): whether the shader should use nearest-neighbor
                texture interpolation. Defaults to True.
    use_texture_alpha (bool, optional): whether to use the alpha channel in the texture
                                        for transparency (only relevant for textured Meshes).
                                        Defaults to True.
    is_billboard (bool, optional): draw this Mesh as a billboard (i.e. always facing the user)
                                rather than rotating with the rest of the world. Defaults to False.
    is_label (bool, optional): this mesh will be treated specially as a label. Not for public use.
                            Defaults to False.

Returns:
    Mesh: a reference to the new Mesh object
"""


Scene.simulate_key_presses = simulate_key_presses
Scene.simulate_key_presses.__doc__ = """Simulates one or more keypresses to fire off events in the ScenePic UI at startup, e.g. video playback.

Args:
    self (Scene): self reference
    canvas_id (Union[Canvas3D, Canvas2D, str]): ID of the canvas on which to fire off this
                                                event.
    keypresses: the key values
"""

Scene._repr_javascript_ = _repr_javascript_



from typing import List


class DropDownMenu:
    """Represents a ScenePic DropDownMenu UI component."""

    @property
    def title(self) -> str:
        """The title of the dropdown menu."""

    @property
    def items(self) -> List[str]:
        """The items in the dropdown menu."""

    @property
    def selection(self) -> str:
        """The currently selected item."""

    def set_disabled(self, index: int, disabled: bool) -> None:
        """Sets the disabled flag on an option.

        Args:
            index (int): index of the menu item
            disabled (bool): whether it should be disabled
        """

    def clear_script(self) -> None:
        """Call this if you are dumping a ScenePic script in multiple parts.

        Use this for client-server interactions, or when chunking a large
        visualization into multiple files.
        """



class UIParameters:
    """Per-Canvas3D parameters of the ScenePic user interface."""

    def __init__(self, pointer_alt_key_multiplier: float,
                 pointer_rotation_speed: float,
                 mouse_wheel_translation_speed: float,
                 layer_dropdown_visibility: str):
        """Constructor.

        Args:
            pointer_alt_key_multiplier (float, optional): when the Alt key is pressed, mouse movement
                                                          is scaled by this factor. Defaults to 0.2.
            pointer_rotation_speed (float, optional): linear scale factor of pointer translation
                                                      into camera rotation. Defaults to 0.01.
            mouse_wheel_translation_speed (float, optional): linear scale factor of mouse wheel input
                                                             into camera translation. Defaults to 0.005.
            layer_dropdown_visibility (str, optional): CSS visibility setting for the layer dropdown menu.
                                                       Defaults to "visible".
        """

    @property
    def pointer_alt_key_multiplier(self) -> float:
        """When the Alt key is pressed, mouse movement is scaled by this factor"""

    @property
    def pointer_rotation_speed(self) -> float:
        """Linear scale factor of pointer translation into camera rotation."""

    @property
    def mouse_wheel_translation_speed(self) -> float:
        """Linear scale factor of mouse wheel input into camera."""

    @property
    def layer_dropdown_visibility(self) -> float:
        """The CSS visibility setting for the layer dropdown menu."""



class TextPanel:
    """Represents a ScenePic TextPanel UI component."""

    @property
    def title(self) -> str:
        """The title of the text panel."""

    @property
    def text(self) -> str:
        """The text within the text panel."""

    @property
    def input_text(self) -> str:
        """Text for the optional input textbox."""

    def clear_script(self) -> None:
        """Call this if you are dumping a ScenePic script in multiple parts.

        Use this for client-server interactions, or when chunking a large
        visualization into multiple files.
        """



class AudioTrack:
    """A ScenePic AudioTrack type."""

    def load(self, path: str) -> None:
        """Load an audio file from the disk.

        Args:
            path (str): the path to the audio file
        """

    def load_from_buffer(self, data: bytes, ext: str) -> None:
        """Load audio from a pre-existing data buffer and extension.

        Args:
            self (AudioTrack): self reference
            data (bytes): the binary data for the audio
            ext (str): the file extension (e.g. mp3, ogg, wav)
        """

    @property
    def video_id(self) -> str:
        """A unique identifier for the audio track."""

    @property
    def data(self) -> bytes:
        """The binary audio data."""
